local array = include( "modules/array" )
local util = include( "modules/util" )
local mathutil = include( "modules/mathutil" )
local cdefs = include( "client_defs" )
local simdefs = include( "sim/simdefs" )
local simquery = include( "sim/simquery" )
local simfactory = include( "sim/simfactory" )
local inventory = include( "sim/inventory" )
local unitdefs = include( "sim/unitdefs" )
local propdefs = include("sim/unitdefs/propdefs")
local usable_item = include( "sim/units/usable_item" )
local mission_util = include( "sim/missions/mission_util" )

---------------------------------------------------------------------------------------------
-- Local helpers

local INCIDENT_NAME = "LOVE LETTER"

local function canPathFunction( cellquery, unit, startcell, endcell )
	if not simquery.canPathBetween( cellquery, unit, startcell, endcell ) then
		return false
	end
	local dx, dy = endcell.x - startcell.x, endcell.y - startcell.y
	if dx == 0 or dy == 0 then
		local dir = simquery.getDirectionFromDelta( dx, dy )
		local exit = startcell.exits[ dir ]
		if exit and not exit.locked then
			return true
		end
	end
	return false
end

local function canSpawnAt(sim,cell)
	if simquery.cellHasTag(sim, cell, "noguard") then        
		return false
	end
	if simquery.cellHasTag(sim, cell, "entry") then        
		return false
	end
	if simquery.cellHasTag(sim, cell, "exit") then        
		return false
	end
	if simquery.cellHasTag(sim, cell, "guard_spawn") then        
		return false
	end
	if simquery.cellHasTag(sim, cell, "door_front") then        
		return false
	end
	if cell.exitID then
		return false
	end
	if cell.impass == nil or cell.impass <= 0 then
		local ok = false
		for i, dir in pairs(simdefs.DIR_SIDES) do
			local dx, dy = simquery.getDeltaFromDirection(dir)
			local otherCell = sim:getCell(cell.x+dx,cell.y+dy)
			if otherCell and otherCell.impass and otherCell.impass > 0 and canPathFunction(sim,nil,cell,otherCell) and sim:nextRand() > 0.5 then
				ok = true
				break
			end
		end
		if not ok then
			return false
		end
	end
	if cell.procgenRoom then
		if cell.procgenRoom.tags.objective then
			return false
		end
		if cell.procgenRoom.tags.entry then
			return false
		end
		if cell.procgenRoom.tags.exit then
			return false
		end
		if cell.procgenRoom.tags.exit_vault then
			return false
		end
	end
	if cell.units and #cell.units > 0 then
		return false
	end
	if cell.isSolid then
		return false
	end
	for i, dir in pairs(simdefs.DIR_SIDES) do
		local dx, dy = simquery.getDeltaFromDirection(dir)
		local otherCell = sim:getCell(cell.x+dx,cell.y+dy)
		if otherCell and (otherCell.impass == nil or otherCell.impass <= 0) and canPathFunction(sim,nil,cell,otherCell) then
			return true
		end
	end
	return false
end

local findSpawnPlaces = function(sim)
	local candidates = {}
	local cell
	sim:forEachCell(function(_cell)
		if _cell.deployID then
			cell = _cell
		end
	end)
	local cells = simquery.floodFill( sim, nil, cell, 999, nil, canPathFunction )
	for i,_cell in pairs(cells) do
		if canSpawnAt(sim,_cell) then
			table.insert(candidates,_cell)
		end
	end
	return candidates
end

local function hook( script, sim )
	local cells = findSpawnPlaces(sim)
	
	--local cells = sim:getCells( "letter_spawn" )
    if cells == nil or #cells < 2 then
		log:write( INCIDENT_NAME .. " : no valid cells" )
		return
	end

    --local cell, targetCell = table.remove( cells ), table.remove( cells )
	local cell = table.remove(cells,sim:nextRand(#cells))
	local targetCell = table.remove(cells,sim:nextRand(#cells))
	while(#cells > 0 and mathutil.distSqr2d(cell.x,cell.y,targetCell.x,targetCell.y) < 50) do
		targetCell = table.remove(cells,sim:nextRand(#cells))
	end
	
	log:write( INCIDENT_NAME .. " : <%d, %d>, <%d, %d>", cell.x, cell.y, targetCell.x, targetCell.y )

	local letter = usable_item.spawnItem( sim, "READ LETTER" )
	sim:warpUnit( letter, cell )
	local letterProp = simfactory.createUnit( propdefs.letter_prop, sim )
	sim:spawnUnit( letterProp )
	sim:warpUnit( letterProp, cell )

	local _, unit = script:waitFor( mission_util.PC_USED( letter ))

   	local txt = 
    {
        string.format( "%s notices a crumpled piece of paper left behind. ", unit:getName() ),
        "It seems to be a love letter.\n\n",
        "Dearest,\nHere is the secret key to my heart: a310d202eff9ab3a. This decodes the document on my desk you were asking about. ",
        "I've come to realize that corporate security is no reason to put boundaries on our trust. Please forgive me.",
    }
	mission_util.showNeutralResult( sim, INCIDENT_NAME, table.concat(txt) )

    sim:warpUnit( letter, nil )
    sim:despawnUnit( letter )
    sim:warpUnit( letterProp, nil )
    sim:despawnUnit( letterProp )

    local intel = usable_item.spawnItem( sim, "PICKUP INTEL" )
    sim:warpUnit( intel, targetCell )
	local intelProp = simfactory.createUnit( propdefs.intel_prop, sim )
	sim:spawnUnit( intelProp )
	sim:warpUnit( intelProp, targetCell )
sim:dispatchEvent( simdefs.EV_UNIT_REFRESH, { unit = intel} ) 
sim:dispatchEvent( simdefs.EV_UNIT_REFRESH, { unit = intelProp} ) 
    unit:getPlayerOwner():glimpseCell( sim, targetCell )
    sim:dispatchEvent( simdefs.EV_CAM_PAN, { targetCell.x, targetCell.y } )	
	
	local _, unit = script:waitFor( mission_util.PC_USED( intel ))
    sim:warpUnit( intel, nil )
    sim:despawnUnit( intel )
    sim:warpUnit( intelProp, nil )
    sim:despawnUnit( intelProp )

    mission_util.showGoodResult( sim, INCIDENT_NAME, "You find the encoded document mentioned in the letter. You scan it through to Central and pass along the secret key for later decoding." )
	mission_util.giftItem( sim, unit, "item_corpIntel" )
end

---------------------------------------------------------------------------------------------
-- Begin!

local incident = class()

function incident:init( scriptMgr, sim )
	self.scriptMgr = scriptMgr
	scriptMgr:addHook( INCIDENT_NAME, hook, false )
end

return incident
